/* cw - play "continuous wave" beeps with a nice envelope
 * pesco 2021
 */

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <sndio.h>
#include <assert.h>
#include <err.h>
#include <unistd.h>	/* sleep() */
#include "cw.h"

/* sndio structures */
static struct sio_hdl *hdl;
static struct sio_par par;
static size_t framesz, bufsz;
static int16_t *buf;

/* envelope parameters */
static int rise = 4;			/* ramp up (attack) time [ms] */
static int fall = 4;			/* ramp down (decay) time [ms] */
static rampfun_t *risefun = ramp_icos;	/* shape of attack envelope */
static rampfun_t *fallfun = ramp_icos;	/* shape of decay envelope */

static int lowtime, hightime;		/* (in)audible time included in ramps */
static int inbeep;			/* did we last beep()? */

/* possible values for the envelope functions */
struct shape { const char *name; rampfun_t *fun; } shapes[] = {
	{ "lin", ramp_lin },	/* linear ramp */
	{ "sqr", ramp_sqr },	/* square function (parabola) */
	{ "sin", ramp_sin },	/* quarter sine wave */
	{ "icos", ramp_icos },	/* inverted cosine */
	{ "exp", ramp_exp },	/* exponential ramp */
	{ NULL, NULL }
};


/* getenv helpers */

int
getenvnum(const char *name, const char *unit, int *p)
{
	char *val;
	int x;

	if ((val = getenv(name)) != NULL) {
		if ((x = atoi(val)) > 0)
			*p = x;
		else
			warnx("%s invalid, defaulting to %d %s",
			    name, *p, unit);
	}

	return *p;
}

rampfun_t *
getenvshape(const char *name, rampfun_t **p)
{
	struct shape *sh;
	char *val;

	if ((val = getenv(name)) != NULL) {
		for (sh = shapes; sh->name != NULL; sh++)
			if (strcmp(val, sh->name) == 0)
				return (*p = sh->fun);

		/* val not found, find and report default setting */
		for (sh = shapes; sh->name != NULL && sh->fun != *p; sh++);
		warnx("%s invalid, defaulting to %s", name, sh->name);
	}

	return *p;
}


/* envelope shapes */

double
ramp_lin(double k)
{
	return k;
}

double
ramp_sqr(double k)
{
	return k * k;
}

double
ramp_sin(double k)
{
	return sin(M_PI_2 * k);
}

double
ramp_icos(double k)
{
	return 1 - cos(M_PI_2 * k);
}

double
ramp_exp(double k)
{
	return exp(1 - 1 / k);
}


/* generate n sine wave samples starting at sample number x */
void
sample_sin(int f, int x, int n)
{
	int16_t y;	/* sample value */
	int i, j;

	for (i = 0; i < n; i++) {
		y = INT16_MAX * sin(2 * M_PI * f * (x + i) / par.rate);

		/* write to all channels */
		for (j = 0; j < par.pchan; j++)
			buf[i * par.pchan + j] = y;
	}
}

/* apply amplitude envelope */
void
envelope(int nramp, int x, int n, int total)
{
	double k;	/* envelope koefficient */
	int i, j;

	if (x >= nramp && x + n <= total - nramp)
		return;

	for (i = 0; i < n; i++) {
		if (x + i < nramp) {
			k = (0.5 + x + i) / nramp;		/* (0, 1) */
			k = risefun(k);
		} else if (x + i >= total - nramp) {
			k = (total - (0.5 + x + i)) / nramp;	/* (1, 0) */
			k = fallfun(k);
		} else
			k = 1.0;

		/* apply to all channels */
		for (j = 0; j < par.pchan; j++)
			buf[i * par.pchan + j] *= k;
	}
}

/* generate and output n samples starting at sample number x */
void
sample_out(int f, int x, int n, int total, int nramp)
{
	assert(n <= bufsz);
	sample_sin(f, x, n);
	envelope(nramp, x, n, total);
	sio_write(hdl, buf, n * framesz);
}

/* play a beep of the desired frequency and length */
void
beep(int f, int len)
{
	int n;		/* samples left to play */
	int x;		/* sample position */
	int total, nrise, nfall;

	len -= hightime;		/* audible time in the ramps */

	total = par.rate * len / 1000;
	nrise = par.rate * rise / 1000;
	nfall = par.rate * fall / 1000;

	n = total;
	x = 0;
	while (n >= bufsz) {
		sample_out(f, x, bufsz, total, nrise);
		x += bufsz;
		n -= bufsz;
	}
	if (n > 0)
		sample_out(f, x, n, total, nfall);

	inbeep = 1;
}

/* play a pause of the desired length */
void
silence(int len)
{
	int n;

	memset(buf, 0, bufsz * framesz);

	if (inbeep)
		len -= lowtime;		/* inaudible time in the ramps */
	n = par.rate * len / 1000;
	while (n >= bufsz) {
		sio_write(hdl, buf, bufsz * framesz);
		n -= bufsz;
	}
	if (n > 0)
		sio_write(hdl, buf, n * framesz);

	inbeep = 0;
}

void
cwinit(void)
{
	/* open sound device and negotiate parameters */
	hdl = sio_open(SIO_DEVANY, SIO_PLAY, 0);
	if (hdl == NULL)
		errx(1, "could not open sound device");
	sio_initpar(&par);
	if (!sio_setpar(hdl, &par))
		errx(1, "sio_setpar failed");
	par.bits = 16;
	par.sig = 1;		/* signed samples */
	par.le = SIO_LE_NATIVE;
	par.pchan = 1;
	if (!sio_getpar(hdl, &par))
		errx(1, "sio_getpar failed");;
	if (par.bits != 16)
		errx(1, "got %d bits per sample instead of 16", par.bits);
	assert(par.bps == sizeof(int16_t));
	if (par.sig != 1)
		errx(1, "got unsigned samples instead of signed");

	/* allocate our buffer */
	framesz = par.bps * par.pchan;
	bufsz = (par.appbufsz + par.round - 1) / par.round * par.round;
	if ((buf = malloc(bufsz * framesz)) == NULL)
		err(1, "malloc");
}

void
cwstart()
{
	getenvnum("CW_ATTACK", "ms", &rise);
	fall = rise;	/* default */
	getenvnum("CW_DECAY", "ms", &fall);
	getenvshape("CW_ATTACK_SHAPE", &risefun);
	getenvshape("CW_DECAY_SHAPE", &fallfun);

	/*
	 * These two parameters are supposed to account for the fact that the
	 * low end of an envelope is inaudible while the high end is already at
	 * full volume. So where does silence end and beep begin?
	 *
	 * The 'lowtime' variable represents time we consider spent in the
	 * inaudible portion of an attack/decay ramp. Conversely, 'hightime'
	 * represents the rest we consider "high" (audible) in attack/decay.
	 */
	lowtime = (rise + fall) / 2;
	hightime = rise + fall - lowtime;

	sio_start(hdl);
}

void
cwstop(void)
{
	sleep(1);	// XXX time to drain buffer
	sio_stop(hdl);
}
